from .base import TgsObject, TgsProp, TgsEnum, todo_func
from .properties import Value, MultiDimensional, GradientColors, ShapeProperty
from .helpers import Transform


def load_shape(lottiedict):
    shapes = {
        'sh': Shape,
        'rc': Rect,
        'el': Ellipse,
        'sr': Star,
        'fl': Fill,
        'gf': GradientFill,
        'gs': GradientStroke,
        'st': Stroke,
        'mm': Merge,
        'tm': Trim,
        'gr': Group,
        'rp': Repeater,
        'tr': TransformShape,
        # RoundedCorners? mentioned but not defined
    }
    return shapes[lottiedict["ty"]].load(lottiedict)


class Rect(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("direction", "d", float, False),
        TgsProp("type", "ty", str, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("size", "s", MultiDimensional, False),
        TgsProp("rounded_corners", "r", Value, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = 0
        # Shape content type.
        self.type = 'rc'
        # Rect's position
        self.position = MultiDimensional([0, 0])
        # Rect's size
        self.size = MultiDimensional([0, 0])
        # Rect's rounded corners
        self.rounded_corners = Value()


class StarType(TgsEnum):
    Star = 1
    Polygon = 2


class Star(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("direction", "d", float, False),
        TgsProp("type", "ty", str, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("inner_radius", "ir", Value, False),
        TgsProp("inner_roundness", "is", Value, False),
        TgsProp("outer_radius", "or", Value, False),
        TgsProp("outer_roundness", "os", Value, False),
        TgsProp("rotation", "r", Value, False),
        TgsProp("points", "pt", Value, False),
        TgsProp("star_type", "sy", StarType, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = 0
        # Shape content type.
        self.type = 'sr'
        # Star's position
        self.position = MultiDimensional([0, 0]) # MultiDimensional, MultiDimensionalKeyframed
        # Star's inner radius. (Star only)
        self.inner_radius = Value() # Value, ValueKeyframed
        # Star's inner roundness. (Star only)
        self.inner_roundness = Value() # Value, ValueKeyframed
        # Star's outer radius.
        self.outer_radius = Value() # Value, ValueKeyframed
        # Star's outer roundness.
        self.outer_roundness = Value() # Value, ValueKeyframed
        # Star's rotation.
        self.rotation = Value() # Value, ValueKeyframed
        # Star's number of points.
        self.points = Value(5) # Value, ValueKeyframed
        # Star's type. Polygon or Star.
        self.star_type = StarType.Star


class Ellipse(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("direction", "d", float, False),
        TgsProp("type", "ty", str, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("size", "s", MultiDimensional, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = 1
        # Shape content type.
        self.type = 'el'
        # Ellipse's position
        self.position = MultiDimensional([0, 0])
        # Ellipse's size
        self.size = MultiDimensional([0, 0])


class Shape(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("direction", "d", float, False),
        TgsProp("type", "ty", str, False),
        TgsProp("vertices", "ks", ShapeProperty, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # After Effect's Direction. Direction how the shape is drawn. Used for trim path for example.
        self.direction = 0
        # Shape content type.
        self.type = 'sh'
        # Shape's vertices
        self.vertices = ShapeProperty()


class Group(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("number_of_properties", "np", float, False),
        TgsProp("shapes", "it", load_shape, True),
        TgsProp("property_index", "ix", int, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'gr'
        # Group number of properties. Used for expressions.
        self.number_of_properties = None
        # Group list of items
        self.shapes = [TransformShape()]
        self.property_index = None

    def add_shape(self, shape):
        self.shapes.insert(-1, shape)
        return shape

    @property
    def transform(self):
        return self.shapes[-1]


class Fill(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("color", "c", MultiDimensional, False),
    ]

    def __init__(self, color=[1, 1, 1]):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'fl'
        # Fill Opacity
        self.opacity = Value(100)
        # Fill Color
        self.color = MultiDimensional(color)


class GradientType(TgsEnum):
    Linear = 1
    Radial = 2


class GradientFill(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("start_point", "s", MultiDimensional, False),
        TgsProp("end_point", "e", MultiDimensional, False),
        TgsProp("gradient_type", "t", GradientType, False),
        TgsProp("highlight_length", "h", Value, False),
        TgsProp("highlight_angle", "a", Value, False),
        TgsProp("colors", "g", GradientColors, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'gf'
        # Fill Opacity
        self.opacity = Value(100)
        # Gradient Start Point
        self.start_point = MultiDimensional([0, 0])
        # Gradient End Point
        self.end_point = MultiDimensional([0, 0])
        # Gradient Type
        self.gradient_type = GradientType.Linear
        # Gradient Highlight Length. Only if type is Radial
        self.highlight_length = Value()
        # Highlight Angle. Only if type is Radial
        self.highlight_angle = Value()
        # Gradient Colors
        self.colors = GradientColors()


class LineJoin(TgsEnum):
    Miter = 1
    Round = 2
    Bevel = 3


class LineCap(TgsEnum):
    Butt = 1
    Round = 2
    Square = 3


class Stroke(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("line_cap", "lc", float, False),
        TgsProp("line_join", "lj", float, False),
        TgsProp("miter_limit", "ml", float, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("width", "w", Value, False),
        TgsProp("color", "c", MultiDimensional, False),
    ]

    def __init__(self, color=[0, 0, 0], width=1):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'st'
        # Stroke Line Cap
        self.line_cap = LineCap.Round
        # Stroke Line Join
        self.line_join = LineJoin.Round
        # Stroke Miter Limit. Only if Line Join is set to Miter.
        self.miter_limit = 0
        # Stroke Opacity
        self.opacity = Value(100)
        # Stroke Width
        self.width = Value(width)
        # Stroke Color
        self.color = MultiDimensional(color)


class GradientStroke(TgsObject):
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("start_point", "s", MultiDimensional, False),
        TgsProp("end_point", "e", MultiDimensional, False),
        TgsProp("gradient_type", "t", GradientType, False),
        TgsProp("highlight_length", "h", Value, False),
        TgsProp("highlight_angle", "a", Value, False),
        TgsProp("colors", "g", GradientColors, False),
        TgsProp("stroke_width", "w", Value, False),
        TgsProp("line_cap", "lc", float, False),
        TgsProp("line_join", "lj", float, False),
        TgsProp("miter_limit", "ml", float, False),
    ]

    def __init__(self, stroke_width=1):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'gs'
        # Stroke Opacity
        self.opacity = Value(100)
        # Gradient Start Point
        self.start_point = MultiDimensional([0, 0])
        # Gradient End Point
        self.end_point = MultiDimensional([0, 0])
        # Gradient Type
        self.gradient_type = GradientType.Linear
        # Gradient Highlight Length. Only if type is Radial
        self.highlight_length = Value()
        # Highlight Angle. Only if type is Radial
        self.highlight_angle = Value()
        # Gradient Colors
        self.colors = GradientColors()
        # Gradient Stroke Width
        self.stroke_width = Value(stroke_width)
        # Gradient Stroke Line Cap
        self.line_cap = LineCap.Round
        # Gradient Stroke Line Join
        self.line_join = LineJoin.Round
        # Gradient Stroke Miter Limit. Only if Line Join is set to Miter.
        self.miter_limit = 0


class TransformShape(TgsObject):
    _props = [
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),

        TgsProp("anchor_point", "a", MultiDimensional, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("scale", "s", MultiDimensional, False),
        TgsProp("rotation", "r", Value, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("skew", "sk", Value, False),
        TgsProp("skew_axis", "sa", Value, False),
    ]

    def __init__(self):
        self.name = None
        self.type = 'tr'
        # Transform Anchor Point
        self.anchor_point = MultiDimensional([0, 0])
        # Transform Position
        self.position = MultiDimensional([0, 0])
        # Transform Scale
        self.scale = MultiDimensional([100, 100])
        # Transform Rotation
        self.rotation = Value(0)
        # Transform Opacity
        self.opacity = Value(100)
        # Transform Skew
        self.skew = Value(0)
        # Transform Skew Axis
        self.skew_axis = Value(0)


class Trim(TgsObject): # TODO check
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("start", "s", Value, False),
        TgsProp("end", "e", Value, False),
        TgsProp("offset", "o", Value, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'tm'
        # Trim Start.
        self.start = Value() # Value, ValueKeyframed
        # Trim End.
        self.end = Value() # Value, ValueKeyframed
        # Trim Offset.
        self.offset = Value() # Value, ValueKeyframed


class Composite(TgsEnum):
    Above = 1
    Below = 2


class Repeater(TgsObject): # TODO check
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("copies", "c", Value, False),
        TgsProp("offset", "o", Value, False),
        TgsProp("composite", "m", Composite, False),
        TgsProp("transform", "tr", Transform, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'rp'
        # Number of Copies
        self.copies = Value() # Value, ValueKeyframed
        # Offset of Copies
        self.offset = Value() # Value, ValueKeyframed
        # Composite of copies
        self.composite = Composite.Above
        # Transform values for each repeater copy
        self.transform = Transform()


class Round(TgsObject): # TODO check
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("radius", "r", Value, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type.
        self.type = 'rd'
        # Rounded Corner Radius
        self.radius = Value() # Value, ValueKeyframed


class Merge(TgsObject): # TODO check
    _props = [
        #TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", str, False),
        TgsProp("merge_mode", "mm", float, False),
    ]

    def __init__(self):
        # After Effect's Match Name. Used for expressions.
        #self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = None
        # Shape content type. THIS FEATURE IS NOT SUPPORTED. It's exported because if you export it, they will come.
        self.type = 'mm'
        # Merge Mode
        self.merge_mode = 0



from .base import TgsObject, todo_func, TgsProp
from .properties import Value, MultiDimensional

load_effect = todo_func


class FillEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", load_effect, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 21
        # Effect List of properties.
        self.effects = [] # PointEffect, DropDownEffect, ColorEffect, DropDownEffect, SliderEffect, SliderEffect, SliderEffect


class StrokeEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", load_effect, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 22
        # Effect List of properties.
        self.effects = [] # ColorEffect, CheckboxEffect, CheckboxEffect, ColorEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, DropDownEffect, DropDownEffect


class DropDownEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", Value, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 7
        # Effect value.
        self.value = Value() # Value, ValueKeyframed


class TritoneEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", todo_func, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 23
        # Effect List of properties.
        self.effects = [] # ColorEffect, ColorEffect, ColorEffect, SliderEffect


class GroupEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", load_effect, True),
        TgsProp("enabled", "en", float, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 5
        # Effect List of properties.
        self.effects = []
        # Enabled AE property value
        self.enabled = 0


class ColorEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", MultiDimensional, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 2
        # Effect value.
        self.value = MultiDimensional() # MultiDimensional, MultiDimensionalKeyframed


class ProLevelsEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", todo_func, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 23
        # ffect List of properties.
        self.effects = [] # DropDownEffect, NoValueEffect, NoValueEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, NoValueEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, NoValueEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, NoValueEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, NoValueEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect, SliderEffect


class AngleEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", Value, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions. NOT USED. EQUALS SLIDER.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 1
        # Effect value.
        self.value = Value() # Value, ValueKeyframed


class SliderEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", Value, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 0
        # Effect value.
        self.value = Value() # Value, ValueKeyframed


class CheckBoxEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", Value, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 7
        # Effect value.
        self.value = Value() # Value, ValueKeyframed


class PointEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", MultiDimensional, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 2
        # Effect value.
        self.value = []


class TintEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("effects", "ef", todo_func, True),
    ]

    def __init__(self):
        # Effect Index. Used for expressions.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 20
        # Effect List of properties.
        self.effects = [] # ColorEffect, ColorEffect, SliderEffect


class LayerEffect(TgsObject): # TODO check
    _props = [
        TgsProp("effect_index", "ix", float, False),
        TgsProp("match_name", "mn", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("type", "ty", float, False),
        TgsProp("value", "v", Value, False),
    ]

    def __init__(self):
        # Effect Index. Used for expressions. NOT USED. EQUALS SLIDER.
        self.effect_index = 0
        # After Effect's Match Name. Used for expressions.
        self.match_name = ""
        # After Effect's Name. Used for expressions.
        self.name = ""
        # Effect type.
        self.type = 0
        # Effect value.
        self.value = Value()



from . import animation, base, effects, enums, helpers, layers, shapes, sources
from .animation import Animation
from .layers import *
from .shapes import *
from .properties import Bezier

__all__ = [
    "animation", "base", "effects", "enums", "helpers", "layers", "shapes", "sources"
    "Animation",
    "NullLayer", "TextLayer", "ShapeLayer", "ImageLayer", "PreCompLayer", "SolidLayer",
    "Rect", "Fill", "Trim", "Repeater", "GFill", "Stroke", "Round", "Shape",
    "TransformShape", "Group", "Star", "Ellipse", "Merge", "GStroke",
    "Bezier"
]



from io import StringIO
from difflib import SequenceMatcher
from ..exporters import prettyprint


def difflines_str(a, b, widtha=None, widthb=None):
    lines_a = a.splitlines()
    lines_b = b.splitlines()
    ia = 0
    ib = 0

    if widtha is None:
        widtha = max(map(len, lines_a))
    if widthb is None:
        widthb = max(map(len, lines_b))

    for ja, jb, size in SequenceMatcher(None, lines_a, lines_b, False).get_matching_blocks():
        sideprinter(lines_a[ia:ja], lines_b[ib:jb], widtha, widthb, "\x1b[31m>", "=", "<\x1b[m")
        ia = ja+size
        ib = jb+size
        sideprinter(lines_a[ja:ia], lines_b[jb:ib], widtha, widthb, "\x1b[m ", "|", " \x1b[m")


def sideprinter(left, right, widtha=40, widthb=40, prefix="", infix=" | ", suffix=""):
    if len(left) > len(right):
        right += [""] * (len(left) - len(right))
    else:
        left += [""] * (len(right) - len(left))
    for l, r in zip(left, right):
        print("".join([prefix, l[:widtha].ljust(widtha), infix, r[:widthb].ljust(widthb), suffix]))


def difflines(a, b, widtha=None, widthb=None):
    ioa = StringIO()
    prettyprint(a, ioa)
    iob = StringIO()
    prettyprint(b, iob)
    difflines_str(ioa.getvalue(), iob.getvalue(), widtha, widthb)



#!/usr/bin/env python3

import json
import sys
import os
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.utils.linediff import difflines
from tgs.objects import Animation


a1 = Animation.load(json.load(open(sys.argv[1])))
a2 = Animation.load(json.load(open(sys.argv[2])))
difflines(a1, a2)



#!/usr/bin/env python3

import gzip
import json
import sys
import codecs

inf = sys.stdin
outf = sys.stdout

if len(sys.argv) > 1 and sys.argv[1] != "-":
    inf = open(sys.argv[1], "r")

if len(sys.argv) > 2 and sys.argv[2] != "-":
    outf = sys.argv[2]


data = json.load(inf)
data["tgs"] = 1

with gzip.open(outf, "wb") as fil:
    json.dump(data, codecs.getwriter('utf-8')(fil))




import re
import math
from xml.etree import ElementTree
from .. import objects
from ..utils.nvector import NVector


ns_map = {
    "dc": "http://purl.org/dc/elements/1.1/",
    "cc": "http://creativecommons.org/ns#",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "svg": "http://www.w3.org/2000/svg",
    "sodipodi": "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd",
    "inkscape": "http://www.inkscape.org/namespaces/inkscape",
}

for n, u in ns_map.items():
    ElementTree.register_namespace(n, u)

element_parsers = {}

color_table = {
    "black": [0, 0, 0, 1],
    "red": [1, 0, 0, 1],
    "lime": [0, 1, 0, 1],
    "yellow": [1, 1, 0, 1],
    "blue": [0, 0, 1, 1],
    "magenta": [1, 0, 1, 1],
    "cyan": [0, 1, 1, 1],
    "white": [1, 1, 1, 1],

    "green": [0, 0.8, 0, 1],
    "purple": [0.5, 0, 0.5, 1],
    "lightgrey": [0.8, 0.8, 0.8, 1],

    "pink": [1, 0.5, 0.5, 1],
}

nocolor = {"none", "transparent"}


def element_parser(name):
    def deco(func):
        element_parsers[name] = func
        return func
    return deco


def _qualified(ns, name):
    return "{%s}%s" % (ns_map[ns], name)


def _simplified(name):
    for k, v in ns_map.items():
        name = name.replace("{%s}" % v, k+":")
    return name


def _unqualified(name):
    return name.split("}")[-1]


def _parse_color(color):
    if re.match("^#[0-9a-fA-F]{6}$", color):
        return [int(color[1:3], 16) / 0xff, int(color[3:5], 16) / 0xff, int(color[5:7], 16) / 0xff, 1]
    if re.match("^#[0-9a-fA-F]{3}$", color):
        return [int(color[1], 16) / 0xf, int(color[2], 16) / 0xf, int(color[3], 16) / 0xf, 1]

    match = re.match("^rgba\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9.eE]+)\s*\)$", color)
    if match:
        return [int(match[1])/255, int(match[2])/255, int(match[3])/255, float(match[4])]

    match = re.match("^rgb\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*([0-9]+)\s*\)$", color)
    if match:
        return [int(match[1])/255, int(match[2])/255, int(match[3])/255, 1]

    # TODO more formats hsv() etc
    return color_table[color]


def _parse_transform(element, dest_trans):
    itcx = _qualified("inkscape", "transform-center-x")
    if itcx in element.attrib:
        cx = float(element.attrib[itcx])
        cy = float(element.attrib[_qualified("inkscape", "transform-center-y")])
        dest_trans.anchor_point.value = [cx, cy]

    if "transform" not in element.attrib:
        return

    for t in re.finditer("([a-zA-Z]+)\s*\(([^\)]*)\)", element.attrib["transform"]):
        name = t[1]
        params = list(map(float, t[2].strip().replace(",", " ").split()))
        if name == "translate":
            dest_trans.position.value = [
                dest_trans.position.value[0] + params[0],
                dest_trans.position.value[1] + (params[1] if len(params) > 1 else 0),
            ]
        elif name == "scale":
            xfac = params[0]
            dest_trans.scale.value[0] = (dest_trans.scale.value[0] / 100 * xfac) * 100
            yfac = params[1] if len(params) > 1 else xfac
            dest_trans.scale.value[1] = (dest_trans.scale.value[1] / 100 * yfac) * 100
        elif name == "rotate":
            ang = params[0]
            x = y = 0
            if len(params) > 2:
                x = params[1]
                y = params[2]
            dest_trans.anchor_point.value = [x, y]
            dest_trans.rotation.value = ang
        elif name == "skewX":
            dest_trans.skew.value = -params[0]
            dest_trans.skew_axis.value = 0
        elif name == "skewY":
            dest_trans.skew.value = params[0]
            dest_trans.skew_axis.value = 90


def _add_shapes(element, shapes, shape_parent):
    # TODO inherit style
    if "style" in element.attrib:
        style = dict(map(
            lambda x: map(lambda y: y.strip(), x.split(":")),
            element.attrib["style"].split(";"))
        )
    else:
        style = {}

    group = objects.Group()
    group.name = element.attrib.get("id")

    shape_parent.shapes.insert(0, group)
    for shape in shapes:
        group.add_shape(shape)

    # TODO opacity:

    stroke_color = style.get("stroke", element.attrib.get("stroke", "none"))
    if stroke_color not in nocolor:
        stroke = objects.Stroke()
        group.add_shape(stroke)
        color = _parse_color(stroke_color)
        stroke.color.value = color[:3]
        stroke.opacity.value = color[3] * 100
        stroke.width.value = float(style.get("stroke-width", 1))
        linecap = style.get("stroke-linecap")
        if linecap == "round":
            stroke.line_cap = objects.shapes.LineCap.Round
        elif linecap == "butt":
            stroke.line_cap = objects.shapes.LineCap.Butt
        elif linecap == "square":
            stroke.line_cap = objects.shapes.LineCap.Square
        linejoin = style.get("stroke-linejoin")
        if linejoin == "round":
            stroke.line_cap = objects.shapes.LineJoin.Round
        elif linejoin == "bevel":
            stroke.line_cap = objects.shapes.LineJoin.Bevel
        elif linejoin in {"miter", "arcs", "miter-clip"}:
            stroke.line_cap = objects.shapes.LineJoin.Miter
        stroke.miter_limit = float(style.get("stroke-miterlimit", 0))

    fill_color = style.get("fill", element.attrib.get("fill", "black"))
    if fill_color not in nocolor:
        color = _parse_color(fill_color)
        fill = group.add_shape(objects.Fill(color[:3]))
        fill.opacity.value = color[3] * 100

    _parse_transform(element, group.transform)

    return group


def _add_shape(element, shape, shape_parent):
    return _add_shapes(element, [shape], shape_parent)


@element_parser("g")
def parse_g(element, shape_parent):
    group = objects.Group()
    shape_parent.add_shape(group)
    _parse_transform(element, group.transform)
    group.name = element.attrib.get(_qualified("inkscape", "label"), element.attrib.get("id"))
    parse_svg_element(element, group)


@element_parser("ellipse")
def parse_ellipse(element, shape_parent):
    ellipse = objects.Ellipse()
    ellipse.position.value = [
        float(element.attrib["cx"]),
        float(element.attrib["cy"])
    ]
    ellipse.size.value = [
        float(element.attrib["rx"]) * 2,
        float(element.attrib["ry"]) * 2
    ]
    _add_shape(element, ellipse, shape_parent)


@element_parser("circle")
def parse_circle(element, shape_parent):
    ellipse = objects.Ellipse()
    ellipse.position.value = [
        float(element.attrib["cx"]),
        float(element.attrib["cy"])
    ]
    r = float(element.attrib["r"]) * 2
    ellipse.size.value = [r, r]
    _add_shape(element, ellipse, shape_parent)


@element_parser("rect")
def parse_ellipse(element, shape_parent):
    rect = objects.Rect()
    w = float(element.attrib["width"])
    h = float(element.attrib["height"])
    rect.position.value = [
        float(element.attrib["x"]) + w / 2,
        float(element.attrib["y"]) + h / 2
    ]
    rect.size.value = [w, h]
    _add_shape(element, rect, shape_parent)


@element_parser("line")
def parse_line(element, shape_parent):
    line = objects.Shape()
    line.vertices.value.add_point([
        float(element.attrib["x1"]),
        float(element.attrib["y1"])
    ])
    line.vertices.value.add_point([
        float(element.attrib["x2"]) * 2,
        float(element.attrib["y2"]) * 2
    ])
    _add_shape(element, line, shape_parent)


@element_parser("polyline")
def parse_line(element, shape_parent):
    line = objects.Shape()
    points = element.attrib["points"].split()
    for point in points:
        line.vertices.value.add_point(
            list(map(float, point.split(",")))
        )
    _add_shape(element, line, shape_parent)


class PathDParser:
    def __init__(self, d_string):
        self.path = objects.properties.Bezier()
        self.paths = [self.path]
        self.p = NVector(0, 0)
        self.la = None
        self.la_type = None
        self.tokens = list(map(self.d_subsplit, re.findall("[a-zA-Z]|[-+.0-9eE,]+", d_string)))
        self.add_p = True
        self.implicit = "M"

    def d_subsplit(self, tok):
        if tok.isalpha():
            return tok
        if "," in tok:
            return NVector(*map(float, tok.split(",")))
        return float(tok)

    def next_token(self):
        if self.tokens:
            self.la = self.tokens.pop(0)
            if isinstance(self.la, str):
                self.la_type = 0
            elif isinstance(self.la, NVector):
                self.la_type = 2
            else:
                self.la_type = 1
        else:
            self.la = None
        return self.la

    def parse(self):
        self.next_token()
        while self.la:
            if self.la_type == 0:
                parser = "_parse_" + self.la
                self.next_token()
                getattr(self, parser)()
            else:
                parser = "_parse_" + self.implicit
                getattr(self, parser)()
            self.next_token()

    def _push_path(self):
        self.path = objects.properties.Bezier()
        self.paths.append(self.path)

    def _parse_M(self):
        if self.la_type != 2:
            self.next_token()
            return
        self.p = self.la
        self.implicit = "L"
        if not self.add_p:
            self._push_path()
        self.add_p = True

    def _parse_m(self):
        if self.la_type != 2:
            self.next_token()
            return
        self.p += self.la
        self.implicit = "l"
        if not self.add_p:
            self._push_path()
        self.add_p = True

    def _rpoint(self, point, rel=None):
        return (point - (rel or self.p)).to_list() if point is not None else [0, 0]

    def _do_add_p(self, outp=None):
        if self.add_p:
            self.path.add_point(self.p.to_list(), [0, 0], self._rpoint(outp))
            self.add_p = False
        elif outp:
            rp = NVector(*self.path.vertices[-1])
            self.path.out_point[-1] = self._rpoint(outp, rp)

    def _parse_L(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        self.p = self.la
        self.path.add_point(self.p.to_list(), NVector(0, 0), NVector(0, 0))
        self.implicit = "L"

    def _parse_l(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        self.p += self.la
        self.path.add_point(self.p.to_list(), [0, 0], [0, 0])
        self.implicit = "l"

    def _parse_H(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[0] = self.la
        self.path.add_point(self.p.to_list(), [0, 0], [0, 0])
        self.implicit = "H"

    def _parse_h(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[0] += self.la
        self.path.add_point(self.p.to_list(), [0, 0], [0, 0])
        self.implicit = "h"

    def _parse_V(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[1] = self.la
        self.path.add_point(self.p.to_list(), [0, 0], [0, 0])
        self.implicit = "V"

    def _parse_v(self):
        if self.la_type != 1:
            self.next_token()
            return
        self._do_add_p()
        self.p[1] += self.la
        self.path.add_point(self.p.to_list(), [0, 0], [0, 0])
        self.implicit = "v"

    def _parse_C(self):
        if self.la_type != 2:
            self.next_token()
            return
        pout = self.la
        self._do_add_p(pout)
        pin = self.next_token()
        self.p = self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "C"

    def _parse_c(self):
        if self.la_type != 2:
            self.next_token()
            return
        pout = self.p + self.la
        self._do_add_p(pout)
        pin = self.p + self.next_token()
        self.p += self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "c"

    def _parse_S(self):
        if self.la_type != 2:
            self.next_token()
            return
        pin = self.la
        self._do_add_p()
        handle = NVector(*self.path.in_point[-1])
        self.path.out_point[-1] = (-handle).to_list()
        self.p = self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "S"

    def _parse_s(self):
        if self.la_type != 2:
            self.next_token()
            return
        pin = self.la + self.p
        self._do_add_p()
        handle = NVector(*self.path.in_point[-1])
        self.path.out_point[-1] = (-handle).to_list()
        self.p += self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "s"

    def _parse_Q(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        pin = self.la
        self.p = self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "Q"

    def _parse_q(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        pin = self.p + self.la
        self.p += self.next_token()
        self.path.add_point(
            self.p.to_list(),
            (pin - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "q"

    def _parse_T(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        handle = -(NVector(*self.path.in_point[-1]) - self.p) + self.p
        self.p = self.la
        self.path.add_point(
            self.p.to_list(),
            (handle - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "T"

    def _parse_t(self):
        if self.la_type != 2:
            self.next_token()
            return
        self._do_add_p()
        handle = -NVector(*self.path.in_point[-1]) + self.p
        self.p += self.la
        self.path.add_point(
            self.p.to_list(),
            (handle - self.p).to_list(),
            [0, 0]
        )
        self.implicit = "t"

    def _parse_A(self):
        if self.la_type != 2:
            self.next_token()
            return
        r = self.la
        xrot = self.next_token()
        large = self.next_token()
        sweep = self.next_token()
        dest = self.next_token()
        self._do_arc(r[0], r[1], xrot, large, sweep, dest)
        self.implicit = "A"

    def _arc_matrix_mul(self, phi, x, y, sin_mul=1):
        c = math.cos(phi)
        s = math.sin(phi) * sin_mul

        xr = c * x - s * y
        yr = s * x + c * y
        return xr, yr

    def _arc_angle(self, u, v):
        arg = math.acos(max(-1, min(1, u.dot(v) / (u.length * v.length))))
        if u[0] * v[1] - u[1] * v[0] < 0:
            return -arg
        return arg

    def _arc_point(self, c, r, xangle, t):
        return NVector(
            c[0] + r[0] * math.cos(xangle) * math.cos(t) - r[1] * math.sin(xangle) * math.sin(t),
            c[1] + r[0] * math.sin(xangle) * math.cos(t) + r[1] * math.cos(xangle) * math.sin(t)
        )

    def _arc_derivative(self, c, r, xangle, t):
        return NVector(
            -r[0] * math.cos(xangle) * math.sin(t) - r[1] * math.sin(xangle) * math.cos(t),
            -r[0] * math.sin(xangle) * math.sin(t) + r[1] * math.cos(xangle) * math.cos(t)
        )

    def _arc_alpha(self, step):
        return math.sin(step) * (math.sqrt(4+3*math.tan(step/2)**2) - 1) / 3

    def _do_arc(self, rx, ry, xrot, large, sweep, dest):
        self._do_add_p()
        if self.p == dest:
            return
        rx = abs(rx)
        ry = abs(ry)

        if rx == 0 or ry == 0:
            # Straight line
            self.p = dest
            self.path.add_point(
                self.p.to_list(),
                [0, 0],
                [0, 0]
            )
            return

        x1 = self.p[0]
        y1 = self.p[1]
        x2 = dest[0]
        y2 = dest[1]
        phi = math.pi * xrot / 180

        tx = (x1 - x2) / 2
        ty = (y1 - y2) / 2
        x1p, y1p = self._arc_matrix_mul(phi, tx, ty, -1)

        cr = x1p ** 2 / rx**2 + y1p**2 / ry**2
        if cr > 1:
            s = math.sqrt(cr)
            rx *= s
            ry *= s

        dq = rx**2 * y1p**2 + ry**2 * x1p**2
        pq = (rx**2 * ry**2 - dq) / dq
        cpm = math.sqrt(max(0, pq))
        if large == sweep:
            cpm = -cpm
        cp = NVector(cpm * rx * y1p / ry, -cpm * ry * x1p / rx)
        c = NVector(*self._arc_matrix_mul(phi, cp[0], cp[1])) + NVector((x1+x2)/2, (y1+y2)/2)
        theta1 = self._arc_angle(NVector(1, 0), NVector((x1p - cp[0]) / rx, (y1p - cp[1]) / ry))
        deltatheta = self._arc_angle(
            NVector((x1p - cp[0]) / rx, (y1p - cp[1]) / ry),
            NVector((-x1p - cp[0]) / rx, (-y1p - cp[1]) / ry)
        ) % (2*math.pi)

        if not sweep and deltatheta > 0:
            deltatheta -= 2*math.pi
        elif sweep and deltatheta < 0:
            deltatheta += 2*math.pi

        r = NVector(rx, ry)
        angle1 = theta1
        angle_left = deltatheta
        step = math.pi / 2
        sign = -1 if theta1+deltatheta < angle1 else 1

        self._do_add_p()
        # We need to fix the first handle
        firststep = min(angle_left, step) * sign
        alpha = self._arc_alpha(firststep)
        q1 = self._arc_derivative(c, r, phi, angle1) * alpha
        self.path.out_point[-1] = q1.to_list()
        # Then we iterate until the angle has been completed
        while abs(angle_left) > step / 2:
            lstep = min(angle_left, step)
            step_sign = lstep * sign
            angle2 = angle1 + step_sign

            alpha = self._arc_alpha(step_sign)
            p2 = self._arc_point(c, r, phi, angle2)
            q2 = -self._arc_derivative(c, r, phi, angle2) * alpha

            self.path.add_smooth_point(
                p2.to_list(),
                q2.to_list(),
            )
            angle1 = angle2
            angle_left -= lstep

        self.p = dest

    def _parse_a(self):
        if self.la_type != 2:
            self.next_token()
            return
        r = self.la
        xrot = self.next_token()
        large = self.next_token()
        sweep = self.next_token()
        dest = self.p + self.next_token()
        self._do_arc(r[0], r[1], xrot, large, sweep, dest)
        self.implicit = "a"

    def _parse_Z(self):
        #self.path.close()
        self._push_path()

    def _parse_z(self):
        #self.path.close()
        self._push_path()


@element_parser("path")
def parse_path(element, shape_parent):
    d_parser = PathDParser(element.attrib.get("d", ""))
    d_parser.parse()
    paths = []
    for path in d_parser.paths:
        p = objects.Shape()
        p.vertices.value = path
        paths.append(p)
    _add_shapes(element, paths, shape_parent)


def parse_svg_element(element, shape_parent):
    for child in element:
        tag = _unqualified(child.tag)
        if tag in element_parsers:
            element_parsers[tag](child, shape_parent)


def parse_svg_etree(etree, *args, **kwargs):
    animation = objects.Animation(*args, **kwargs)
    svg = etree.getroot()
    if "width" in svg.attrib and "height" in svg.attrib:
        animation.width = int(svg.attrib["width"])
        animation.height = int(svg.attrib["height"])
    else:
        _, _, animation.width, animation.height = map(int, svg.attrib["viewBox"].split(" "))
    animation.name = svg.attrib.get(_qualified("sodipodi", "docname"), svg.attrib.get("id"))
    layer = objects.ShapeLayer()
    animation.add_layer(layer)
    parse_svg_element(svg, layer)
    return animation


def parse_svg_file(file, *args, **kwargs):
    return parse_svg_etree(ElementTree.parse(file), *args, **kwargs)



from .base import TgsObject, TgsProp, PseudoBool
from .effects import load_effect
from .helpers import Transform
from .shapes import load_shape
from .enums import BlendMode


def load_layer(lottiedict):
    layers = {
        0: PreCompLayer,
        1: SolidLayer,
        2: ImageLayer,
        3: NullLayer,
        4: ShapeLayer,
        5: TextLayer,
    }
    return layers[lottiedict["ty"]].load(lottiedict)


class NullLayer(TgsObject): # TODO check
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        TgsProp("css_class", "cl", str, False),
        TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", str, False),
        TgsProp("effects", "ef", load_effect, True),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
    ]

    def __init__(self):
        # Type of layer: Null.
        self.type = 3
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = 0
        # Parsed layer name used as html class on SVG/HTML renderer
        self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = 0
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = 0
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = None
        # List of Effects
        self.effects = [] # IndexEffect
        # Layer Time Stretching
        self.stretch = 0
        # Layer Parent. Uses ind of parent.
        self.parent = None


class TextLayer(TgsObject): # TODO check
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", float, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        TgsProp("css_class", "cl", str, False),
        TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", float, False),
        #TgsProp("has_masks", "hasMask", float, False),
        #TgsProp("masks_properties", "masksProperties", Mask, True),
        TgsProp("effects", "ef", PseudoBool, False),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
        TgsProp("text_data", "t", float, False),
    ]

    def __init__(self):
        # Type of layer: Text.
        self.type = 0
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # Blend Mode
        self.blend_mode = BlendMode.default()
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = 0
        # Parsed layer name used as html class on SVG/HTML renderer
        self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = 0
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = 0
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = 0
        # Boolean when layer has a mask. Will be deprecated in favor of checking masksProperties.
        self.has_masks = 0
        # List of Masks
        self.masks_properties = [] # Mask
        # Auto-Orient along path AE property.
        self.effects = False
        # Layer Time Stretching
        self.stretch = 0
        # Layer Parent. Uses ind of parent.
        self.parent = None
        # Text Data
        self.text_data = None


class ShapeLayer(TgsObject):
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", float, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        #TgsProp("css_class", "cl", str, False),
        #TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", str, False),
        #TgsProp("has_masks", "hasMask", float, False),
        #TgsProp("masks_properties", "masksProperties", Mask, True),
        TgsProp("effects", "ef", load_effect, True),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
        TgsProp("shapes", "shapes", load_shape, True),
    ]

    def __init__(self):
        # Type of layer: Shape.
        self.type = 4
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # Blend Mode
        self.blend_mode = BlendMode.default()
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = None
        ## Parsed layer name used as html class on SVG/HTML renderer
        #self.css_class = ""
        ## Parsed layer name used as html id on SVG/HTML renderer
        #self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = None
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = None
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = None
        # Boolean when layer has a mask. Will be deprecated in favor of checking masksProperties.
        self.has_masks = 0
        # List of Masks
        self.masks_properties = [] # Mask
        # List of Effects
        self.effects = None # IndexEffect
        # Layer Time Stretching
        self.stretch = 1
        # Layer Parent. Uses ind of parent.
        self.parent = None
        # Shape list of items
        self.shapes = [] # Shape, Rect, Ellipse, Star, Fill, GFill, GStroke, Stroke, Merge, Trim, Group, RoundedCorners, Repeater

    def add_shape(self, shape):
            self.shapes.append(shape)
            return shape

class ImageLayer(TgsObject): # TODO check
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", float, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        TgsProp("css_class", "cl", str, False),
        TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", float, False),
        #TgsProp("has_masks", "hasMask", float, False),
        #TgsProp("masks_properties", "masksProperties", Mask, True),
        TgsProp("effects", "ef", load_effect, True),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
        TgsProp("reference_id", "refId", str, False),
    ]

    def __init__(self):
        # Type of layer: Image.
        self.type = 2
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # Blend Mode
        self.blend_mode = BlendMode.default()
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = 0
        # Parsed layer name used as html class on SVG/HTML renderer
        self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = 0
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = 0
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = 0
        # Boolean when layer has a mask. Will be deprecated in favor of checking masksProperties.
        self.has_masks = 0
        # List of Masks
        self.masks_properties = [] # Mask
        # List of Effects
        self.effects = [] # IndexEffect
        # Layer Time Stretching
        self.stretch = 0
        # Layer Parent. Uses ind of parent.
        self.parent = None
        # id pointing to the source image defined on 'assets' object
        self.reference_id = ""


class PreCompLayer(TgsObject): # TODO check
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", float, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        TgsProp("css_class", "cl", str, False),
        TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", float, False),
        #TgsProp("has_masks", "hasMask", float, False),
        #TgsProp("masks_properties", "masksProperties", Mask, True),
        TgsProp("effects", "ef", load_effect, True),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
        TgsProp("reference_id", "refId", str, False),
        TgsProp("time_remapping", "tm", float, False),
    ]

    def __init__(self):
        # Type of layer: Precomp.
        self.type = 0
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # Blend Mode
        self.blend_mode = BlendMode.default()
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = 0
        # Parsed layer name used as html class on SVG/HTML renderer
        self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = 0
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = 0
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = 0
        # Boolean when layer has a mask. Will be deprecated in favor of checking masksProperties.
        self.has_masks = 0
        # List of Masks
        self.masks_properties = [] # Mask
        # List of Effects
        self.effects = [] # IndexEffect
        # Layer Time Stretching
        self.stretch = 0
        # Layer Parent. Uses ind of parent.
        self.parent = None
        # id pointing to the source composition defined on 'assets' object
        self.reference_id = ""
        # Comp's Time remapping
        self.time_remapping = Value()


class SolidLayer(TgsObject): # TODO check
    _props = [
        TgsProp("type", "ty", float, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", float, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", float, False),
        TgsProp("css_class", "cl", str, False),
        TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", float, False),
        #TgsProp("has_masks", "hasMask", float, False),
        #TgsProp("masks_properties", "masksProperties", Mask, True),
        TgsProp("effects", "ef", PseudoBool, False),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", float, False),
        TgsProp("solid_color", "sc", str, False),
        TgsProp("solid_height", "sh", float, False),
        TgsProp("solid_width", "sw", float, False),
    ]

    def __init__(self):
        # Type of layer: Solid.
        self.type = 0
        # Transform properties
        self.transform = Transform()
        # Auto-Orient along path AE property.
        self.auto_orient = False
        # Blend Mode
        self.blend_mode = BlendMode.default()
        # 3d layer flag
        self.threedimensional = False
        # Layer index in AE. Used for parenting and expressions.
        self.index = 0
        # Parsed layer name used as html class on SVG/HTML renderer
        self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        self.layer_html_id = ""
        # In Point of layer. Sets the initial frame of the layer.
        self.in_point = 0
        # Out Point of layer. Sets the final frame of the layer.
        self.out_point = 0
        # Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        # After Effects Layer Name. Used for expressions.
        self.name = 0
        # Boolean when layer has a mask. Will be deprecated in favor of checking masksProperties.
        self.has_masks = 0
        # List of Masks
        self.masks_properties = [] # Mask
        # Auto-Orient along path AE property.
        self.effects = False
        # Layer Time Stretching
        self.stretch = 0
        # Layer Parent. Uses ind of parent.
        self.parent = None
        # Color of the solid in hex
        self.solid_color = ""
        # Height of the solid.
        self.solid_height = 0
        # Width of the solid.
        self.solid_width = 0



#!/usr/bin/env python3

import gzip
import json
import sys

inf = sys.stdin.buffer
outf = sys.stdout

if len(sys.argv) > 1 and sys.argv[1] != "-":
    inf = sys.argv[1]

if len(sys.argv) > 2 and sys.argv[2] != "-":
    outf = open(sys.argv[2], "w")


with gzip.open(inf, "rb") as unzipped:
    json.dump(
        json.load(unzipped),
        outf,
        indent=4
    )



from .base import TgsObject, TgsProp, todo_func
from .properties import MultiDimensional, Value


class Transform(TgsObject):
    _props = [
        TgsProp("anchor_point", "a", MultiDimensional, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("scale", "s", MultiDimensional, False),
        TgsProp("rotation", "r", Value, False),
        TgsProp("opacity", "o", Value, False),
        #TgsProp("position_x", "px", Value, False),
        #TgsProp("position_y", "py", Value, False),
        #TgsProp("position_z", "pz", Value, False),
        TgsProp("skew", "sk", Value, False),
        TgsProp("skew_axis", "sa", Value, False),
    ]

    def __init__(self):
        # Transform Anchor Point
        self.anchor_point = MultiDimensional([0, 0, 0]) # MultiDimensional, MultiDimensionalKeyframed
        # Transform Position
        self.position = MultiDimensional([0, 0]) # MultiDimensional, MultiDimensionalKeyframed
        # Transform Scale
        self.scale = MultiDimensional([100, 100, 100]) # MultiDimensional, MultiDimensionalKeyframed
        # Transform Rotation
        self.rotation = Value(0) # Value, ValueKeyframed
        # Transform Opacity
        self.opacity = Value(100) # Value, ValueKeyframed
        # Transform Position X
        #self.position_x = Value() # Value, ValueKeyframed
        ## Transform Position Y
        #self.position_y = Value() # Value, ValueKeyframed
        ## Transform Position Z
        #self.position_z = Value() # Value, ValueKeyframed
        # Transform Skew
        self.skew = Value() # Value, ValueKeyframed
        # Transform Skew Axis
        self.skew_axis = Value() # Value, ValueKeyframed


class Mask(TgsObject): # TODO check
    _props = [
        TgsProp("inverted", "inv", float, False),
        TgsProp("name", "nm", str, False),
        TgsProp("points", "pt", todo_func, False),
        TgsProp("opacity", "o", todo_func, False),
        TgsProp("mode", "mode", str, False),
    ]

    def __init__(self):
        # Inverted Mask flag
        self.inverted = None
        # Mask name. Used for expressions and effects.
        self.name = ""
        # Mask vertices
        self.points = ShapeProperty() # ShapeProperty, ShapePropertyKeyframed
        # Mask opacity.
        self.opacity = Value(100) # Const, ConstKeyframed
        # Mask mode. Not all mask types are supported.
        self.mode = ""



import enum
import inspect


def _to_dict(v):
    if isinstance(v, Tgs):
        return v.to_dict()
    elif isinstance(v, list):
        return list(map(_to_dict, v))
    elif isinstance(v, (int, float, str, bool)):
        return v
    else:
        raise Exception("Unknown value %r" % v)


class Tgs:
    def to_dict(self):
        raise NotImplementedError

    @classmethod
    def load(cls, lottiedict):
        raise NotImplementedError


class TgsEnum(Tgs, enum.Enum):
    def to_dict(self):
        return self.value

    @classmethod
    def load(cls, lottieint):
        return cls(lottieint)


class PseudoList:
    pass


class TgsConverter:
    def __init__(self, py, lottie):
        self.py = py
        self.lottie = lottie

    def py_to_lottie(self, val):
        return self.lottie(val)

    def lottie_to_py(self, val):
        return self.py(val)


PseudoBool = TgsConverter(bool, int)


class TgsProp:
    def __init__(self, name, lottie, type=float, list=False, cond=None):
        self.name = name
        self.lottie = lottie
        self.type = type
        self.list = list
        self.cond = cond

    def get(self, obj):
        return getattr(obj, self.name)

    def set(self, obj, value):
        return setattr(obj, self.name, value)

    def load_from_parent(self, lottiedict):
        if self.lottie in lottiedict:
            return self.load(lottiedict[self.lottie])
        return None

    def load_into(self, lottiedict, obj):
        if self.cond and not self.cond(lottiedict):
            return
        self.set(obj, self.load_from_parent(lottiedict))

    def load(self, lottieval):
        if self.list is PseudoList and isinstance(lottieval, list):
            return self.load_scalar(lottieval[0])
        elif self.list is True:
            return [
                self.load_scalar(it)
                for it in lottieval
            ]
        return self.load_scalar(lottieval)

    def load_scalar(self, lottieval):
        if inspect.isclass(self.type) and issubclass(self.type, Tgs):
            return self.type.load(lottieval)
        elif isinstance(self.type, type) and isinstance(lottieval, self.type):
            return lottieval
        elif isinstance(self.type, TgsConverter):
            return self.type.lottie_to_py(lottieval)
        return self.type(lottieval)

    def to_dict(self, obj):
        val = _to_dict(self.get(obj))
        if self.list is PseudoList:
            val = [val]
        elif isinstance(self.type, TgsConverter):
            return self.type.py_to_lottie(val)
        return val


class TgsObject(Tgs):
    def to_dict(self):
        return {
            prop.lottie: prop.to_dict(self)
            for prop in self._props
            if prop.get(self) is not None
        }

    @classmethod
    def load(cls, lottiedict):
        obj = cls()
        for prop in cls._props:
            prop.load_into(lottiedict, obj)
        return obj

    def find(self, search, propname="name"):
        if getattr(self, propname, None) == search:
            return self
        for prop in self._props:
            v = prop.get(self)
            if isinstance(v, TgsObject):
                found = v.find(search, propname)
                if found:
                    return found
            elif isinstance(v, list) and v and isinstance(v[0], TgsObject):
                for obj in v:
                    found = obj.find(search, propname)
                    if found:
                        return found
        return None


class Index:
    def __init__(self):
            self._i = -1

    def __next__(self):
            self._i += 1
            return self._i


def todo_func(x):
    raise NotImplementedError()



from .base import TgsEnum


class TestBased(TgsEnum):
    Characters = 1
    CharacterExcludingSpaces = 2
    Words = 3
    Lines = 4

    @classmethod
    def default(cls):
        return cls.Characters


class TextShape(TgsEnum):
    Square = 1
    RampUp = 2
    RampDown = 3
    Triangle = 4
    Round = 5
    Smooth = 6

    @classmethod
    def default(cls):
        return cls.Square


class TextGrouping(TgsEnum):
    Characters = 1
    Word = 2
    Line = 3
    All = 4

    @classmethod
    def default(cls):
        return cls.Characters


class BlendMode(TgsEnum):
    Normal = 0
    Multiply = 1
    Screen = 2
    Overlay = 3
    Darken = 4
    Lighten = 5
    ColorDodge = 6
    ColorBurn = 7
    HardLight = 8
    SoftLight = 9
    Difference = 10
    Exclusion = 11
    Hue = 12
    Saturation = 13
    Color = 14
    Luminosity = 15

    @classmethod
    def default(cls):
        return cls.Normal



import operator
import math


def vop(op, a, b):
    return list(map(op, a, b))


class NVector():
    def __init__(self, *components):
        self.components = list(components)

    def __str__(self):
        return str(self.components)

    def __repr__(self):
        return "<NVector %s>" % self

    def __len__(self):
        return len(self.components)

    def to_list(self):
        return list(self.components)

    def __add__(self, other):
        return NVector(*vop(operator.add, self.components, other.components))

    def __sub__(self, other):
        return NVector(*vop(operator.sub, self.components, other.components))

    def __mul__(self, scalar):
        return NVector(*(c * scalar for c in self.components))

    def __truediv__(self, scalar):
        return NVector(*(c / scalar for c in self.components))

    def __iadd__(self, other):
        self.components = vop(operator.add, self.components, other.components)
        return self

    def __isub__(self, other):
        self.components = vop(operator.sub, self.components, other.components)
        return self

    def __imul__(self, scalar):
        self.components = [c * scalar for c in self.components]
        return self

    def __itruediv__(self, scalar):
        self.components = [c / scalar for c in self.components]
        return self

    def __neg__(self):
        return NVector(*(-c for c in self.components))

    def __getitem__(self, key):
        return self.components[key]

    def __setitem__(self, key, value):
        self.components[key] = value

    def __eq__(self, other):
        return self.components == other.components

    @property
    def length(self):
        return math.sqrt(sum(map(lambda x: x**2, self.components)))

    def dot(self, other):
        return sum(map(operator.mul, self.components, other.components))



import math
from functools import reduce
from .base import TgsObject, TgsProp, PseudoList, PseudoBool


class KeyframeBezierPoint(TgsObject):
    _props = [
        TgsProp("time", "x", list=PseudoList),
        TgsProp("value", "y", list=PseudoList),
    ]

    def __init__(self, time=0, value=0):
        self.time = time
        self.value = value


def _set_tangents(self, start, end, end_time):
    self.in_value = KeyframeBezierPoint(
        (end_time - self.time) / 3,
        (end - start) / 3
    )
    self.out_value = KeyframeBezierPoint(
        (end_time - self.time) / 3,
        (end - start) / 3
    )

    time_scale = end_time - self.time
    time_diff = self.time / time_scale
    value_scale = end - start
    if value_scale == 0:
        value_scale = 1e-9
    value_diff = start / value_scale

    self.in_value.time = end_time - self.in_value.time
    self.in_value.value = end - self.in_value.value
    self.in_value.time = abs(self.in_value.time / time_scale - time_diff)
    self.in_value.value = abs(self.in_value.value / value_scale - value_diff)

    self.out_value.time += self.time
    self.out_value.value += start
    self.out_value.time = abs(self.out_value.time / time_scale - time_diff)
    self.out_value.value = abs(self.out_value.value / value_scale - value_diff)


class OffsetKeyframe(TgsObject):
    _props = {
        TgsProp("start", "s", float, True),
        TgsProp("end", "e", float, True),
        TgsProp("time", "t", float, False),
        TgsProp("in_value", "i", KeyframeBezierPoint, False),
        TgsProp("out_value", "o", KeyframeBezierPoint, False),
        TgsProp("in_tan", "ti", float, True),
        TgsProp("out_tan", "to", float, True),
        #TgsProp("h" ,"h"),
    }

    def __init__(self, time=0, start=None, end=None,
                 in_value=None, out_value=None, in_tan=None, out_tan=None):
        # Start value of keyframe segment.
        self.start = start # number
        self.end = end # number
        # Start time of keyframe segment.
        self.time = time
        # Bezier curve interpolation in value.
        self.in_value = in_value
        # Bezier curve interpolation out value.
        self.out_value = out_value
        #self.h = 1 #???
        # In Spatial Tangent. Only for spatial properties. Array of numbers.
        self.in_tan = in_tan
        # Out Spatial Tangent. Only for spatial properties. Array of numbers.
        self.out_tan = out_tan

    def set_tangents(self, end_time):
        to_val = lambda vals: math.sqrt(sum(map(lambda x: x**2, vals)))
        start = to_val(self.start)
        end = to_val(self.end)
        _set_tangents(self, start, end, end_time)


class MultiDimensional(TgsObject):
    _props = [
        TgsProp("value", "k", float, True, lambda l: not l["a"]),
        TgsProp("property_index", "ix", int, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", OffsetKeyframe, True, lambda l: l["a"]),
    ]

    def __init__(self, value=None):
        # Property Value
        self.value = value
        # Property Index. Used for expressions.
        self.property_index = None
        self.animated = False
        # Property Value keyframes
        self.keyframes = None # [OffsetKeyframe]

    def clear_animation(self, value):
        self.value = value
        self.animated = False

    def add_keyframe(self, time, value):
        if not self.animated:
            self.value = None
            self.keyframes = []
            self.animated = True
        else:
            self.keyframes[-1].end = value
            self.keyframes[-1].set_tangents(time)

        self.keyframes.append(OffsetKeyframe(
            time,
            value,
            None,
            None,
            None,
        ))


class GradientColors(TgsObject):
    _props = [
        TgsProp("colors", "k", MultiDimensional),
        TgsProp("count", "p", int),
    ]

    def __init__(self):
        self.colors = MultiDimensional()
        self.count = 0

    def set_colors(self, colors, keyframe=None):
        flat = self._flatten_colors(colors)
        if self.colors.animated and keyframe is not None:
            if keyframe > 1:
                self.colors.keyframes[keyframe-1].end = flat
            self.colors.keyframes[keyframe].start = flat
        else:
            self.colors.clear_animation(flat)
        self.count = len(colors)

    def _flatten_colors(self, colors):
        return reduce(
            lambda a, b: a + b,
            map(
                lambda it: [it[0] / (len(colors)-1)] + it[1],
                enumerate(colors),
            )
        )

    def add_color(self, offset, color, keyframe=None):
        flat = [offset] + color
        if self.colors.animated:
            if keyframe is None:
                for kf in self.colors.keyframes:
                    if kf.start:
                        kf.start += flat
                    if kf.end:
                        kf.end += flat
            else:
                if keyframe > 1:
                    self.colors.keyframes[keyframe-1].end += flat
                self.colors.keyframes[keyframe].start += flat
        else:
            self.colors.value += flat
        self.count += 1

    def add_keyframe(self, time, colors=None):
        self.colors.add_keyframe(time, self._flatten_colors(colors) if colors else [])


class Value(TgsObject):
    _props = [
        TgsProp("value", "k", float, False, lambda l: not l["a"]),
        TgsProp("property_index", "ix", int, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", OffsetKeyframe, True, lambda l: l["a"]),
    ]

    def __init__(self, value=0):
        # Property Value
        self.value = value
        # Property Expression. An AE expression that modifies the value.
        self.expression = None
        # Property Index. Used for expressions.
        self.property_index = None
        self.animated = False
        self.keyframes = None

    def clear_animation(self, value):
        self.value = value
        self.animated = False

    def add_keyframe(self, time, value):
        if not self.animated:
            self.value = None
            self.keyframes = []
            self.animated = True
        else:
            self.keyframes[-1].end = [value]
            self.keyframes[-1].set_tangents(time)

        self.keyframes.append(OffsetKeyframe(
            time,
            [value],
            None,
            None,
            None,
        ))


class Bezier(TgsObject):
    _props = [
        TgsProp("closed", "c", bool, False),
        TgsProp("in_point", "i", list, True),
        TgsProp("out_point", "o", list, True),
        TgsProp("vertices", "v", list, True),
    ]

    def __init__(self):
        # Closed property of shape
        self.closed = False
        # Bezier curve In points. Array of 2 dimensional arrays.
        self.in_point = []
        # Bezier curve Out points. Array of 2 dimensional arrays.
        self.out_point = []
        # Bezier curve Vertices. Array of 2 dimensional arrays.
        self.vertices = []

    def add_point(self, pos, inp=[0, 0], outp=[0, 0]):
        self.vertices.append(pos)
        self.in_point.append(list(inp))
        self.out_point.append(list(outp))
        return self

    def add_smooth_point(self, pos, inp):
        self.add_point(pos, inp, [-inp[0], -inp[1]])
        return self

    def close(self, closed=True):
        self.closed = closed
        return self


class ShapePropKeyframe(TgsObject): # TODO check
    _props = [
        TgsProp("start", "s", Bezier, PseudoList),
        TgsProp("end", "e", Bezier, PseudoList),
        TgsProp("time", "t", float, False),
        TgsProp("in_value", "i", KeyframeBezierPoint, False),
        TgsProp("out_value", "o", KeyframeBezierPoint, False),
    ]

    def __init__(self, time=0, start=None, end=None):
        # Start value of keyframe segment.
        self.start = start
        self.end = end
        # Start time of keyframe segment.
        self.time = time
        # Bezier curve interpolation in value.
        self.in_value = None
        # Bezier curve interpolation out value.
        self.out_value = None

    def set_tangents(self, end_time):
        _set_tangents(self, 0, 0, end_time)


class ShapeProperty(TgsObject):
    _props = [
        TgsProp("value", "k", Bezier, False, lambda l: not l["a"]),
        #TgsProp("expression", "x", str, False),
        TgsProp("property_index", "ix", float, False),
        TgsProp("animated", "a", PseudoBool, False),
        TgsProp("keyframes", "k", ShapePropKeyframe, True, lambda l: l["a"]),
    ]

    def __init__(self):
        # Property Value
        self.value = Bezier()
        # Property Expression. An AE expression that modifies the value.
        #self.expression = None
        # Property Index. Used for expressions.
        self.property_index = None
        # Defines if property is animated
        self.animated = False
        self.keyframes = None

    def clear_animation(self, shape):
        self.value = shape
        self.animated = False

    def add_keyframe(self, time, shape):
        if not self.animated:
            self.value = None
            self.keyframes = []
            self.animated = True
        else:
            self.keyframes[-1].end = shape
            self.keyframes[-1].set_tangents(time)

        self.keyframes.append(ShapePropKeyframe(
            time,
            shape,
            None,
        ))


class DoubleKeyframe(TgsObject): # TODO check
    _props = [
        TgsProp("start", "s", float, False),
        TgsProp("time", "t", float, False),
        TgsProp("in_value", "i", float, False),
        TgsProp("out_value", "o", float, False),
    ]

    def __init__(self):
        # Start value of keyframe segment.
        self.start = 0
        # Start time of keyframe segment.
        self.time = 0
        # Bezier curve interpolation in value.
        self.in_value = None
        # Bezier curve interpolation out value.
        self.out_value = None




from .base import TgsObject, TgsProp, todo_func


class Image(TgsObject): # TODO check
    _props = [
        TgsProp("height", "h", float, False),
        TgsProp("width", "w", float, False),
        TgsProp("id", "id", str, False),
        TgsProp("image_name", "p", str, False),
        TgsProp("image_path", "u", str, False),
    ]

    def __init__(self):
        # Image Height
        self.height = 0
        # Image Width
        self.width = 0
        # Image ID
        self.id = ""
        # Image name
        self.image_name = ""
        # Image path
        self.image_path = ""


class Chars(TgsObject): # TODO check
    _props = [
        TgsProp("character", "ch", str, False),
        TgsProp("font_family", "fFamily", str, False),
        TgsProp("font_size", "size", str, False),
        TgsProp("font_style", "style", str, False),
        TgsProp("width", "w", float, False),
        TgsProp("character_data", "data", float, False),
    ]

    def __init__(self):
        # Character Value
        self.character = ""
        # Character Font Family
        self.font_family = ""
        # Character Font Size
        self.font_size = ""
        # Character Font Style
        self.font_style = ""
        # Character Width
        self.width = 0
        # Character Data
        self.character_data = None


class Precomp(TgsObject): # TODO check
    _props = [
        TgsProp("id", "id", str, False),
        TgsProp("layers", "layers", todo_func, True),
    ]

    def __init__(self):
        # Precomp ID
        self.id = ""
        # List of Precomp Layers
        self.layers = [] # ShapeLayer, SolidLayer, CompLayer, ImageLayer, NullLayer, TextLayer



#!/usr/bin/env python3

import json
import sys
import os
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.exporters import prettyprint
from tgs.objects import Animation


prettyprint(Animation.load(json.load(open(sys.argv[1]))))




#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.parsers.svg import parse_svg_file
from tgs.exporters import export_tgs, export_lottie

parser = argparse.ArgumentParser(
    description="Converts a (static) SVG image into a TGS file"
)
parser.add_argument(
    "infile",
    nargs="?",
    default="-",
    help="Input file"
)
parser.add_argument(
    "outfile",
    nargs="?",
    default="-",
    help="Output file"
)
parser.add_argument(
    "--format",
    "-f",
    default="tgs",
    choices=["tgs", "lottie"],
    help="Output format"
)
parser.add_argument(
    "--framerate",
    default=60,
    type=int,
    help="Output file framerate"
)
parser.add_argument(
    "--frames",
    default=60,
    type=int,
    help="Output file frame count"
)
ns = parser.parse_args()

if ns.infile == "-":
    infile = sys.stdin
else:
    infile = ns.infile

animation = parse_svg_file(infile, ns.frames, ns.framerate)


binary = ns.format == "tgs"
if ns.outfile == "-":
    outfile = sys.stdout
    if binary:
        outfile = outfile.buffer
else:
    outfile = open(ns.outfile, "w" + ("b" if binary else ""))

if ns.format == "lottie":
    export_lottie(animation, outfile)
else:
    export_tgs(animation, outfile)



from .base import TgsObject, TgsProp, PseudoBool, todo_func
from .layers import load_layer


class Animation(TgsObject):
    _props = {
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("frame_rate", "fr", float, False),
        TgsProp("width", "w", int, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("height", "h", int, False),
        TgsProp("version", "v", str, False),
        TgsProp("name", "nm", str, False),
        TgsProp("layers", "layers", load_layer, True),
        TgsProp("assets", "assets", todo_func, True),
        #TgsProp("chars", "chars", Chars, True),
        TgsProp("tgs", "tgs", PseudoBool, False),
    }

    def __init__(self, n_frames=60, framerate=60):
        self.tgs = 1
        # In Point of the Time Ruler. Sets the initial Frame of the animation.
        self.in_point = 0
        # Out Point of the Time Ruler. Sets the final Frame of the animation
        self.out_point = n_frames
        # Frame Rate
        self.frame_rate = framerate
        # Composition Width
        self.width = 512
        # Composition has 3-D layers
        self.threedimensional = False
        # Composition Height
        self.height = 512
        # Bodymovin Version
        self.version = "5.5.2"
        # Composition name
        self.name = None
        # List of Composition Layers
        self.layers = [] # ShapeLayer, SolidLayer, CompLayer, ImageLayer, NullLayer, TextLayer
        # source items that can be used in multiple places. Comps and Images for now.
        self.assets = [] # Image, Precomp
        # source chars for text layers
        #self.chars = [] # Chars


    def add_layer(self, layer):
        self.layers.append(layer)
        if layer.index is None:
            layer.index = len(self.layers)
        if layer.in_point is None:
            layer.in_point = self.in_point
        if layer.out_point is None:
            layer.out_point = self.out_point
        return layer



import setuptools
import os
here = os.path.dirname(os.path.abspath(__file__))


with open(os.path.join(here, "README.md"), "r") as fh:
    long_description = fh.read()


def find_packages(root="tgs"):
    absroot = os.path.join(here, "lib", root)
    paks = []
    for sub in os.listdir(absroot):
        if sub == "__pycache__":
            continue
        lname = os.path.join(root, sub)
        if os.path.isdir(os.path.join(absroot, sub)):
            paks += [lname]
            paks += find_packages(lname)
    return paks


setuptools.setup(
    name="tgs",
    version="0.1.0",
    author="Mattia Basaglia",
    author_email="mattia.basaglia@gmail.com",
    description="A framework to work with lottie / tgs files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/mattia.basaglia/tgs/",
    package_dir={'': 'lib'},
    packages=find_packages(),
    scripts=[
        os.path.join("bin", "lottie2tgs.py"),
        os.path.join("bin", "lottiecat.py"),
        os.path.join("bin", "lottiediff.py"),
        os.path.join("bin", "svg2tgs.py"),
        os.path.join("bin", "tgs2lottie.py"),
    ],
    keywords="telegram stickers tgs lottie svg animation",
    # https://pypi.org/classifiers/
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Graphics",
    ],
)
